package no.vestein.rtree;

import no.vestein.rtree.lsm.*;
import org.apache.commons.cli.*;

public class Main {

  public static void main(String[] arg) {
    try {
      Options options = new Options();
      options.addOption("n", true, "Number of elements");
      options.addOption("b", true, "Buffersize");
      options.addOption("f", true, "Fanout");
      options.addOption("s", true, "Seed for random number generation");

      options.addOption("draw", false, "Draw part of tree.");
      options.addOption("curve", true, "Z or H. Default is Z. (Z-Order / Hilbert)");
      options.addOption("tag", true, "Tag for tagging outputs");

      options.addOption("runs", true, "Number of runs at the highest level");
      options.addOption("levels", true, "Number of levels in tree");
      options.addOption("bt", true, "Number of buffer threads");
      options.addOption("bmt", true, "Number of buffer merge threads");
      options.addOption("mt", true, "Number of merge threads");

      options.addOption("mstrat", true, "Merge strategy");
      options.addOption("rbs", true, "Read Buffer Size");
      options.addOption("wbs", true, "Write Buffer Size");
      options.addOption("rf", true, "Record multiplier. Increases record size");

      options.addOption("out", true, "Output folder");


      CommandLineParser parser = new DefaultParser();
      CommandLine cmd = parser.parse(options, arg);


      int numberOfElements = Integer.parseInt(cmd.getOptionValue("n", "10000"));
      int bufferSize = Integer.parseInt(cmd.getOptionValue("b", String.valueOf((int) Math.pow(2, 16))));
      int fanout = Integer.parseInt(cmd.getOptionValue("f", String.valueOf((int) Math.pow(2, 4))));
      long seed = Long.parseLong(cmd.getOptionValue("s", String.valueOf(1337L)));
      boolean drawOutput = cmd.hasOption("draw");
      String curveAlg = cmd.getOptionValue("curve", "z").toLowerCase();
      String tag = cmd.getOptionValue("tag");
      String out = cmd.getOptionValue("out", "data") + "/";

      int runs = Integer.parseInt(cmd.getOptionValue("runs", String.valueOf(4)));
      int levels = Integer.parseInt(cmd.getOptionValue("levels", String.valueOf(4)));
      int bt = Integer.parseInt(cmd.getOptionValue("bt", String.valueOf(2)));
      int bmt = Integer.parseInt(cmd.getOptionValue("bmt", String.valueOf(2)));
      int mt = Integer.parseInt(cmd.getOptionValue("mt", String.valueOf(4)));

      BitMagic.SpaceCurve curve = new BitMagic.ZOrder();
      if (curveAlg.equals("h")) {
        curve = new BitMagic.Hilbert();
      }

      LSM.MergeStrat mstrat;
      try {
        mstrat = LSM.MergeStrat.valueOf(cmd.getOptionValue("mstrat", "continuous").toUpperCase()) ;
      } catch (IllegalArgumentException e) {
        mstrat = LSM.MergeStrat.CONTINUOUS;
      }

      int readBufferSize = Integer.parseInt(cmd.getOptionValue("rbs", "1048576"));
      int writeBufferSize = Integer.parseInt(cmd.getOptionValue("wbs", "1048576"));
      int rf = Integer.parseInt(cmd.getOptionValue("rf", "1"));

      testLSMTree(
          numberOfElements,
          bufferSize,
          fanout,
          seed,
          drawOutput,
          curve,
          tag,
          out,
          runs,
          levels,
          bt,
          bmt,
          mt,
          mstrat,
          readBufferSize,
          writeBufferSize,
          rf
      );

    } catch (ParseException e) {
      e.printStackTrace();
    }

  }

  public static void testLSMTree(
      int numberOfElements, int bufferSize, int fanout, long seed,
      boolean draw, BitMagic.SpaceCurve curve, String tag, String out,
      int runs,
      int levels,
      int bt,
      int bmt,
      int mt,
      LSM.MergeStrat mstrat,
      int readBufferSize,
      int writeBufferSize,
      int rf
  ) {
    String dataDir = out;
    long start;
    long end;
    Test.cleanFolder(dataDir);

    Test.EntrySet entrySet = new Test.UniformEntrySet(seed, numberOfElements);
//    Test.EntrySet entrySet = new Test.NormalEntrySet(seed, numberOfElements, 32);


    LSM lsm = new LSM(fanout , bufferSize, curve, dataDir, runs, levels, bt, bmt, mt, mstrat, readBufferSize, writeBufferSize, rf);
    start = System.currentTimeMillis();

    while (entrySet.entries.peekFirst() != null) {
      lsm.insert(entrySet.entries.pollFirst());
    }

    lsm.stop();
    end = System.currentTimeMillis();

    Stats.time.set(end - start);

    String meta = String.format("%s--#%d-B%d-F%d-S%d-curve%s-bt%d-bmt%d-mt%d-mergeStrat%s-rbs%d-wbs%d-rf%d",
        tag, numberOfElements, bufferSize, fanout, seed, curve.shortName(), bt, bmt, mt, mstrat, readBufferSize, writeBufferSize, rf);

    Test.writeStats(String.format("%s.stats", meta));

    if (draw) {
      String pngFile = String.format("%s.png", meta);

      Visuals.drawPersisTree(lsm.getComponentManager(2).trees[0], pngFile);
    }

  }

}
