package no.vestein.rtree.lsm;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.nio.ByteBuffer;
import java.util.LinkedList;
import java.util.List;

public class HilbertPersistentTree extends PersistentTree {

  public HilbertPersistentTree(int fanout, int generation, int level, int c, String dataDir) {
    super(fanout, generation, level, c, dataDir);
  }

  @Override
  public synchronized void merge(PersistentTree[] trees, int bufferSize) {
    int n = trees.length;

    IOLSM.NodeFileHeader entryHeader;

    FileInputStream[] streams = new FileInputStream[n];
    int[] inputStatus = new int[n];
    IOLSM.NodeFileHeader[] headers = new IOLSM.NodeFileHeader[n];
    ByteBuffer[] readBuffers = new ByteBuffer[n];
    LinkedList<Entry>[] entryBuffer = new LinkedList[n];

    // Output for leaves
    ByteBuffer writeBuffer = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
    FileOutputStream output;

    try {

      long totalSize = 0;
      for (int i = 0; i < n; i++) {
        streams[i] = new FileInputStream(trees[i].getLeafFile());
        headers[i] = IOLSM.bytesToNodeFileHeader(streams[i].readNBytes(IOLSM.HEADER_SIZE));

//        readBuffers[i] = ByteBuffer.allocate(bufferSize * IOLSM.ENTRY_SIZE);
        readBuffers[i] = ByteBuffer.allocate(IOLSM.READ_BUFFER_SIZE);
        entryBuffer[i] = new LinkedList<>();
        inputStatus[i] = 1;

        totalSize += headers[i].size;
      }

      // Expected tree height
      height = calcHeight(totalSize, fanout);

      // Prepare leaf file
      entryHeader = new IOLSM.NodeFileHeader(totalSize, IOLSM.ENTRY_SIZE, height - 1, IOLSM.NodeFileHeader.ENTRY);
      output = new FileOutputStream(getLevelFileName(height - 1));
      output.write(IOLSM.nodeFileHeaderToBytes(entryHeader));

      // Merge input files
      fillEntryBuffers(n, entryBuffer, streams, readBuffers, inputStatus);

      while (!isAllEmpty(entryBuffer, n)) {
        Entry next = getNext(entryBuffer, n);

        writeBuffer.put(IOLSM.entryToBytes(next));
        if (writeBuffer.remaining() < IOLSM.ENTRY_SIZE) {
          flushBuffer1(output, writeBuffer);
        }

        boolean status = fillEntryBuffers(n, entryBuffer, streams, readBuffers, inputStatus);
      }

      // Flush entry level
      flushBuffer1(output, writeBuffer);
      output.close();

      // Close input streams
      for (int i = 0; i < n; i++) {
        streams[i].close();
      }

      createLeafNodes();

    } catch (Exception e) {
      e.printStackTrace();
    }

  }

  public synchronized void createTree(List<Entry> entries) {
    IOLSM.NodeFileHeader entryHeader;

    // Output for leaves
    ByteBuffer writeBuffer = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
    FileOutputStream output;

    try {
      long totalSize = entries.size();

      // Expected tree height
      height = calcHeight(totalSize, fanout);

      // Prepare leaf file
      entryHeader = new IOLSM.NodeFileHeader(totalSize, IOLSM.ENTRY_SIZE, height - 1, IOLSM.NodeFileHeader.ENTRY);
      output = new FileOutputStream(getLevelFileName(height - 1));
      output.write(IOLSM.nodeFileHeaderToBytes(entryHeader));

      // Merge input files

      Entry next;
      for (int i = 0; i < totalSize; i++) {
        next = entries.get(i);
        writeBuffer.put(IOLSM.entryToBytes(next));
        if (writeBuffer.remaining() < IOLSM.ENTRY_SIZE) {
          flushBuffer1(output, writeBuffer);
        }
      }

      // Flush entry level
      flushBuffer1(output, writeBuffer);
      output.close();

      createLeafNodes();
    } catch (Exception e) {
      e.printStackTrace();
    }

  }

  private void createLeafNodes() {
    FileInputStream entriesFile;
    ByteBuffer inputBuffer;
    LinkedList<Entry> entries;
    IOLSM.NodeFileHeader inputHeader;

    FileOutputStream outputStream;
    ByteBuffer writeBuffer;
    IOLSM.NodeFileHeader outputHeader;

    try {
      entriesFile = new FileInputStream(getLeafFile());
      inputBuffer = ByteBuffer.allocate(IOLSM.READ_BUFFER_SIZE);
      entries = new LinkedList<>();
      inputHeader = IOLSM.bytesToNodeFileHeader(entriesFile.readNBytes(IOLSM.HEADER_SIZE));

      outputStream = new FileOutputStream(getLevelFileName(height - 2));
      writeBuffer = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
      outputHeader = new IOLSM.NodeFileHeader(parentNodesNeeded(inputHeader.size, fanout), IOLSM.NODE_SIZE, height - 2, IOLSM.NodeFileHeader.NODE);

      outputStream.write(IOLSM.nodeFileHeaderToBytes(outputHeader));

      IOLSM.fillBuffer1(entries, entriesFile, inputBuffer);

      int counter = 1;
      Entry next = entries.pollFirst();
      Node tNode = new Node(next.xl, next.xh, next.yl, next.yh, IOLSM.HEADER_SIZE, IOLSM.HEADER_SIZE + IOLSM.ENTRY_SIZE, height - 2);
      while (!entries.isEmpty()) {
        next = entries.pollFirst();

        if (counter >= fanout) {
          writeBuffer.put(IOLSM.nodeToBytes(tNode));
          counter = 0;
          tNode.setData(next);
          tNode.start = tNode.start + tNode.length;
          tNode.length = IOLSM.ENTRY_SIZE;
        } else {
          tNode.updateDimension(next);
          tNode.updateLhv(next.lhv);
          tNode.length += IOLSM.ENTRY_SIZE;
        }
        counter++;

        if (writeBuffer.remaining() < IOLSM.NODE_SIZE) {
          flushBuffer1(outputStream, writeBuffer);
        }

        if (entries.isEmpty()) {
          IOLSM.fillBuffer1(entries, entriesFile, inputBuffer);
        }

      }

      writeBuffer.put(IOLSM.nodeToBytes(tNode));
      flushBuffer1(outputStream, writeBuffer);
      outputStream.close();

      entriesFile.close();

      createNodeLevel(height - 2);

    } catch (Exception e) {
      e.printStackTrace();
    }

  }

  private void createNodeLevel(int readLevel) {

    FileInputStream inputStream;
    ByteBuffer inputBuffer;
    LinkedList<Node> inputNodes;
    IOLSM.NodeFileHeader inputHeader;

    FileOutputStream outputStream;
    ByteBuffer writeBuffer;
    IOLSM.NodeFileHeader outputHeader;

    try {
      inputStream = new FileInputStream(getLevelFileName(readLevel));
      inputBuffer = ByteBuffer.allocate(IOLSM.READ_BUFFER_SIZE);
      inputNodes = new LinkedList<>();
      inputHeader = IOLSM.bytesToNodeFileHeader(inputStream.readNBytes(IOLSM.HEADER_SIZE));

      outputStream = new FileOutputStream(getLevelFileName(readLevel - 1));
      writeBuffer = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
      outputHeader = new IOLSM.NodeFileHeader(parentNodesNeeded(inputHeader.size, fanout), IOLSM.NODE_SIZE, readLevel - 1, IOLSM.NodeFileHeader.NODE);

      outputStream.write(IOLSM.nodeFileHeaderToBytes(outputHeader));

      IOLSM.fillBufferNodes1(inputNodes, inputStream, inputBuffer);

      int counter = 1;
      Node next = inputNodes.pollFirst();
      Node tNode = new Node(next.xl, next.xh, next.yl, next.yh, IOLSM.HEADER_SIZE, IOLSM.HEADER_SIZE + IOLSM.ENTRY_SIZE, height - 2);
      while (!inputNodes.isEmpty()) {
        next = inputNodes.pollFirst();

        if (counter >= fanout) {
          writeBuffer.put(IOLSM.nodeToBytes(tNode));
          counter = 0;
          tNode.setData(next);
          tNode.start = tNode.start + tNode.length;
          tNode.length = IOLSM.ENTRY_SIZE;
        } else {
          tNode.updateDimension(next);
          tNode.updateLhv(next.lhv);
          tNode.length += IOLSM.ENTRY_SIZE;
        }
        counter++;

        if (writeBuffer.remaining() < IOLSM.NODE_SIZE) {
          flushBuffer1(outputStream, writeBuffer);
        }

        if (inputNodes.isEmpty()) {
          IOLSM.fillBufferNodes1(inputNodes, inputStream, inputBuffer);
        }

      }

      writeBuffer.put(IOLSM.nodeToBytes(tNode));
      flushBuffer1(outputStream, writeBuffer);
      outputStream.close();

      inputStream.close();

      if (readLevel - 1 > 0) {
        createNodeLevel(readLevel - 1);
      } else {
        createRootLevel();
      }

    } catch (Exception e) {
      e.printStackTrace();
    }

  }

  private void createRootLevel() {
    FileInputStream inputStream;
    ByteBuffer inputBuffer;
    LinkedList<Node> inputNodes;
    IOLSM.NodeFileHeader inputHeader;

    FileOutputStream outputStream;
    ByteBuffer writeBuffer;
    IOLSM.NodeFileHeader rootHeader = new IOLSM.NodeFileHeader(1, IOLSM.NODE_SIZE, -1, IOLSM.NodeFileHeader.ROOT);

    try {
      inputStream = new FileInputStream(getLevelFileName(0));
      inputBuffer = ByteBuffer.allocate(IOLSM.READ_BUFFER_SIZE);
      inputNodes = new LinkedList<>();
      inputHeader = IOLSM.bytesToNodeFileHeader(inputStream.readNBytes(IOLSM.HEADER_SIZE));

      outputStream = new FileOutputStream(getRootFile());
      writeBuffer = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);

      outputStream.write(IOLSM.nodeFileHeaderToBytes(rootHeader));

      IOLSM.fillBufferNodes1(inputNodes, inputStream, inputBuffer);

      //int counter = 1;
      Node next = inputNodes.pollFirst();
      Node root = new Node(next.xl, next.xh, next.yl, next.yh, IOLSM.HEADER_SIZE, (int) (IOLSM.HEADER_SIZE + IOLSM.NODE_SIZE * inputHeader.size), -1);
      while (!inputNodes.isEmpty()) {
        next = inputNodes.pollFirst();

        root.updateDimension(next);
        root.updateLhv(next.lhv);

        if (inputNodes.isEmpty()) {
          IOLSM.fillBufferNodes1(inputNodes, inputStream, inputBuffer);
        }

      }
      writeBuffer.put(IOLSM.nodeToBytes(root));

      flushBuffer1(outputStream, writeBuffer);
      outputStream.close();
      inputStream.close();

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

}
