package no.vestein.rtree.lsm;

import java.util.List;
import java.util.concurrent.*;
import no.vestein.rtree.lsm.BitMagic.SpaceCurve;

// TODO: file handling for snapshots etc..
// TODO: remove old code and comments
// TODO: better thread handling (akka?)

public class LSM {
  private final Buffer[] buffers;
  private final int BUFFER_SIZE;
  private final int FANOUT;
  private final int runs;
  private final int levels;
  private final int bufferSize;
  private final int bufferThreads;
  public final SpaceCurve spaceCurve;
  public final MergeStrat mergeStrat;

  private ComponentManager[] componentManagers;

  private ConcurrentLinkedQueue<Callable<Boolean>> pendingTasks;
  private ConcurrentLinkedQueue<Future> runningTasks;
  private TaskRunner runner;
  private Thread runnerThread;
  private ExecutorService poolMerge;
  private ExecutorService poolBuffer;
  private ExecutorService poolBufferMerge;
  private ConcurrentLinkedQueue<Entry> inputQueue;
  private String DATA_DIR;


  public LSM(final int fanout,
             final int bufferSize,
             BitMagic.SpaceCurve spaceCurve,
             String dataDir,
             int runs,
             int levels,
             int bufferThreads,
             int bufferMergeThreads,
             int mergeThreads,
             MergeStrat mergeStrat,
             int readBufferSize,
             int writeBufferSize,
             int rf
  ) {
    this.FANOUT = fanout;
    this.BUFFER_SIZE = bufferSize;
    this.spaceCurve = spaceCurve;
    this.DATA_DIR = dataDir;

    this.runs = runs;
    this.levels = levels;
    this.bufferSize = bufferSize;
    this.bufferThreads = bufferThreads;
    this.mergeStrat = mergeStrat;

    IOLSM.READ_BUFFER_SIZE = readBufferSize;
    IOLSM.WRITE_BUFFER_SIZE = writeBufferSize;
    IOLSM.ENTRY_SIZE *= rf;


    runner = new TaskRunner();
    pendingTasks = new ConcurrentLinkedQueue<>();
    runningTasks = new ConcurrentLinkedQueue<>();
    poolMerge = Executors.newFixedThreadPool(mergeThreads);
    poolBuffer = Executors.newFixedThreadPool(bufferThreads);
    poolBufferMerge = Executors.newFixedThreadPool(bufferMergeThreads);

    buffers = new Buffer[bufferThreads];

    initInputQueues();
    initBuffers();
    initComponentManagers();

    initBufferQueue();

    initTaskRunner();

  }

  public void stop() {
    boolean finished = false;
    while (!finished) {
      if (inputQueue.isEmpty()) {
        finished = true;
      }
    }

    while (runningTasks.size() > 0 || pendingTasks.size() > 0) {
      try {
        Thread.sleep(1);
      } catch (InterruptedException e) {

      }
    }

    runnerThread.interrupt();
    poolMerge.shutdownNow();
    poolBuffer.shutdownNow();
    poolBufferMerge.shutdownNow();
  }

  private void initTaskRunner() {
    runnerThread = new Thread(runner);
    runnerThread.start();
//    pool.execute(runner);
  }

  private void initInputQueues() {
    inputQueue = new ConcurrentLinkedQueue<>();
  }

  private void initBufferQueue() {
    for (int i = 0; i < bufferThreads; i++) {
      poolBuffer.execute(new BufferRunner(i));
    }
  }

  private void initBuffers() {
    for (int i = 0; i < bufferThreads; i++) {
      buffers[i] = new Buffer(bufferSize);
    }
  }

  private void initComponentManagers() {
    componentManagers = new ComponentManager[levels];
    for (int i = 0; i < levels; i++) {
      componentManagers[i] = new ComponentManager(FANOUT, i, runs * (int) Math.pow(2, i), poolMerge, DATA_DIR, mergeStrat);
    }
  }

  public void insert(Entry entry) {
    inputQueue.add(entry);
  }

  public ComponentManager getComponentManager(int level) {
    return componentManagers[level];
  }

  // TODO: akka?
  public class TaskRunner implements Runnable {
    private Callable<Boolean> nextTask;

    @Override
    public void run() {
      while (!Thread.currentThread().isInterrupted()) {

        nextTask = pendingTasks.poll();
        if (nextTask != null) {
          try {
            Future<Boolean> f;
            if (nextTask instanceof MergeTaskBuffer) {
              f = poolBufferMerge.submit(nextTask);
            } else {
              f = poolMerge.submit(nextTask);
            }
            runningTasks.add(f);
          } catch (RejectedExecutionException e) {
            e.printStackTrace();
            pendingTasks.add(nextTask);
          }
//          runningTasks.add(pool.submit(nextTask));
        }

        for (Future f : runningTasks) {
          if (f.isDone() || f.isCancelled()) {
            runningTasks.remove(f);
          }
        }

      }
    }
  }

  public class MergeTaskTree implements Callable<Boolean> {
    private final int id;
    private final ComponentManager manager;

    public MergeTaskTree(int id, ComponentManager manager) {
      this.id = id;
      this.manager = manager;
    }

    @Override
    public Boolean call() throws Exception {
      MergeTaskTree mergeTask = new MergeTaskTree(id, componentManagers[manager.level - 1]);
      PersistentTree tree;

      // TODO
      while ((tree = componentManagers[manager.level - 1].getFreeRun(mergeTask)) == null) {
        Thread.sleep(10);
        pendingTasks.add(this);
        return false;
      }
      tree.merge(manager.trees, BUFFER_SIZE);
      manager.updateGeneration();
      manager.runCounter.set(0);
      manager.merging.compareAndSet(true, false);

      return true;
    }

  }

  public class MergeTaskBuffer implements Callable<Boolean> {
    private final int id;
    private List<Entry> entries;
    private final Buffer buffer;

    public MergeTaskBuffer(Buffer buffer, int id/*, int runs, int lower, int upper*/) {
      this.buffer = buffer;
      this.id = id;
    }

    @Override
    public Boolean call() throws Exception {
      try {
        int level = levels - 1;
        entries = buffer.getSorted();

        MergeTaskTree mergeTask = new MergeTaskTree(id, componentManagers[level]);
        PersistentTree tree;

        // TODO
        while ((tree = componentManagers[level].getFreeRun(mergeTask)) == null) {
          Thread.sleep(10);
        }
        tree.createTree(entries);
        return true;
      } catch (Exception e) {
        e.printStackTrace();
      }

      return false;
    }
  }

  public class BufferRunner implements Runnable {
    public final int id;
    private Entry next;

    public BufferRunner(int id/*, int runs, int lower, int upper*/) {
      this.id = id;
    }

    @Override
    public void run() {
      while (!Thread.currentThread().isInterrupted()) {
        next = inputQueue.poll();
        if (next != null) {
          next.lhv = spaceCurve.encode(next.xl, next.yl);
          buffers[id].add(next);
          if (buffers[id].isFull()) {
            pendingTasks.add(new MergeTaskBuffer(buffers[id], id/*, runs, lower, upper*/));
            buffers[id] = new Buffer(bufferSize);
          }
        }

      }
    }


  }
  public enum MergeStrat {
    CONTINUOUS,
    MAPPED,
    PARTLY;

    public PersistentTree createInstance(int fanout, int generation, int level, int c, String dataDir) {
      switch (this) {
        case CONTINUOUS:
          return new StreamPersistentTree(fanout, generation, level, c, dataDir);
        case MAPPED:
          return new MappedPersistentTree(fanout, generation, level, c, dataDir);
        case PARTLY:
          return new HilbertPersistentTree(fanout, generation, level, c, dataDir);
        default:
          return new StreamPersistentTree(fanout, generation, level, c, dataDir);
      }
    }

  }

}
