package no.vestein.rtree.lsm;


import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.LinkedList;
import java.util.List;

public class MappedPersistentTree extends PersistentTree {

  public MappedPersistentTree(int fanout, int generation, int level, int c, String dataDir) {
    super(fanout, generation, level, c, dataDir);
  }

  @Override
  public synchronized void merge(PersistentTree[] trees, int bufferSize) {
    int n = trees.length;

    IOLSM.NodeFileHeader entryHeader;

    int[] inputStatus = new int[n];
    IOLSM.NodeFileHeader[] headers = new IOLSM.NodeFileHeader[n];
    LinkedList<Entry>[] entryBuffer = new LinkedList[n];
    RandomAccessFile[] inputFiles = new RandomAccessFile[n];
    MappedByteBuffer[] inputBuffers = new MappedByteBuffer[n];


    // Output for leaves
    RandomAccessFile outputFile;
    MappedByteBuffer output;

    // Output buffers for node levels
    Node[] tempNodes;
    int[][] parentCounters;
    IOLSM.NodeFileHeader[] nodeOutputHeaders;
    RandomAccessFile[] nodeOutputFiles;
    MappedByteBuffer[] nodeOutputs;

    try {
      long totalSize = 0L;
      byte[] headerBuffer = new byte[IOLSM.HEADER_SIZE];
      for (int i = 0; i < n; i++) {
        inputFiles[i] = new RandomAccessFile(trees[i].getLeafFile(), "r");
        inputFiles[i].read(headerBuffer);
        headers[i] = IOLSM.bytesToNodeFileHeader(headerBuffer);
        inputBuffers[i] = inputFiles[i].getChannel().map(FileChannel.MapMode.READ_ONLY, IOLSM.HEADER_SIZE, headers[i].size * headers[i].elementLength);
        entryBuffer[i] = new LinkedList<>();
        inputStatus[i] = 1;

        totalSize += headers[i].size;
      }

      // Expected tree height
      height = calcHeight(totalSize, fanout);

      // Output initialization
      nodeOutputFiles = new RandomAccessFile[height];
      nodeOutputs = new MappedByteBuffer[height];
      tempNodes = new Node[height];
      parentCounters = new int[height][2];
      nodeOutputHeaders = new IOLSM.NodeFileHeader[height];


      for (int i = height - 1; i > 0; i--) {
        int totalNodes = nodesAtGivenLevel(totalSize, fanout, i - 1);
        nodeOutputFiles[i] = new RandomAccessFile(getLevelFileName(i - 1), "rw");
        nodeOutputs[i] = nodeOutputFiles[i].getChannel().map(FileChannel.MapMode.READ_WRITE, 0, IOLSM.HEADER_SIZE + IOLSM.NODE_SIZE * totalNodes);

        parentCounters[i][0] = fanout;
        parentCounters[i][1] = IOLSM.HEADER_SIZE;

        tempNodes[i] = new Node(Float.MAX_VALUE, -Float.MAX_VALUE, Float.MAX_VALUE, -Float.MAX_VALUE);
        nodeOutputHeaders[i] = new IOLSM.NodeFileHeader(totalNodes, IOLSM.NODE_SIZE, i - 1, IOLSM.NodeFileHeader.NODE);
        nodeOutputs[i].put(IOLSM.nodeFileHeaderToBytes(nodeOutputHeaders[i]));
      }

      // Prepare root file
      nodeOutputFiles[0] = new RandomAccessFile(getRootFile(), "rw");
      nodeOutputs[0] = nodeOutputFiles[0].getChannel().map(FileChannel.MapMode.READ_WRITE, 0, IOLSM.NODE_SIZE);

      parentCounters[0][0] = fanout;
      tempNodes[0] = new Node(Float.MAX_VALUE, -Float.MAX_VALUE, Float.MAX_VALUE, -Float.MAX_VALUE);
      nodeOutputHeaders[0] = new IOLSM.NodeFileHeader(1, IOLSM.NODE_SIZE, -1, IOLSM.NodeFileHeader.ROOT);
      nodeOutputs[0].put(IOLSM.nodeFileHeaderToBytes(nodeOutputHeaders[0]));

      // Prepare leaf file
      entryHeader = new IOLSM.NodeFileHeader(totalSize, IOLSM.ENTRY_SIZE, height - 1, IOLSM.NodeFileHeader.ENTRY);
      outputFile = new RandomAccessFile(getLevelFileName(height - 1), "rw");
      output = outputFile.getChannel().map(FileChannel.MapMode.READ_WRITE, 0, IOLSM.HEADER_SIZE + (totalSize * IOLSM.ENTRY_SIZE));
      output.put(IOLSM.nodeFileHeaderToBytes(entryHeader));


      // Merge input files
      fillEntryBuffersMapped(n, entryBuffer, inputBuffers, inputStatus);

      ByteBuffer entryBuff = ByteBuffer.allocate(IOLSM.ENTRY_SIZE);

      Entry next;
      while (!isAllEmpty(entryBuffer, n)) {
        next = getNext(entryBuffer, n);
        updateParentsMapped(next, height - 1, nodeOutputs, tempNodes, parentCounters);

        output.put(IOLSM.entryToBytes1(next, entryBuff));

        fillEntryBuffersMapped(n, entryBuffer, inputBuffers, inputStatus);
      }

      // Flush entry level
      outputFile.close();
      output = null;

      // Close input streams
      for (int i = 0; i < n; i++) {
        inputFiles[i].close();
        inputBuffers[i] = null;
      }

      // Force update upwards
      for (int i = height - 2; i > 0; i--) {
        tempNodes[i].updateDimension(tempNodes[i + 1]);
        tempNodes[i].updateLhv(tempNodes[i + 1].lhv);
        tempNodes[i].length += IOLSM.NODE_SIZE;
      }

      // Flush node levels
      for (int i = 0; i < height; i++) {
        nodeOutputFiles[i].close();
        nodeOutputs[i] = null;
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  @Override
  public void createTree(List<Entry> entries) {

  }

  // fix this
  private void updateParentsMapped(Entry entry, int level, ByteBuffer[] nodeOutputs, Node[] tempNodes, int[][] parentCounters) throws IOException {
    if (parentCounters[level][0] > 0) {
      tempNodes[level].updateDimension(entry);
      tempNodes[level].updateLhv(entry.lhv);
      tempNodes[level].length += IOLSM.ENTRY_SIZE;
      parentCounters[level][0]--;
      return;
    }

    nodeOutputs[level].put(IOLSM.nodeToBytes(tempNodes[level]));

    int ups = level;
    while (--ups >= 0) {
      if (parentCounters[ups][0] > 0) {
        tempNodes[ups].updateDimension(tempNodes[ups + 1]);
        tempNodes[ups].updateLhv(tempNodes[ups + 1].lhv);
        tempNodes[ups].length += IOLSM.NODE_SIZE;
        parentCounters[ups][0]--;
        break;
      }

      nodeOutputs[ups].put(IOLSM.nodeToBytes(tempNodes[ups]));
    }

    parentCounters[level][0] = fanout - 1;
    tempNodes[level].setData(entry);
    tempNodes[level].start += tempNodes[level].length;
    tempNodes[level].length = IOLSM.ENTRY_SIZE;

    while (++ups < level) {
      parentCounters[ups][0] = fanout - 1;
      tempNodes[ups].setData(tempNodes[ups + 1]);
      tempNodes[ups].start += tempNodes[ups].length;
      tempNodes[ups].length = IOLSM.NODE_SIZE;
    }
  }

  private boolean fillEntryBuffersMapped(int n, LinkedList<Entry>[] entryBuffers, ByteBuffer[] buffers, int[] inputStatus) throws IOException {
    byte[] b = new byte[IOLSM.ENTRY_SIZE];
    for (int i = 0; i < n; i++) {
      if (entryBuffers[i].isEmpty() && inputStatus[i] > 0) {
        inputStatus[i] = IOLSM.fillBufferMapped(entryBuffers[i], buffers[i], b);
      }
    }

    boolean status = true;
    for (int i = 0; i < n; i++) {
      if (inputStatus[i] > 0) {
        status = false;
      }
    }
    return status;
  }

}
