package no.vestein.rtree.lsm;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.LinkedList;
import java.util.List;

public abstract class PersistentTree {
  protected final int fanout;
  protected final int level;
  protected final String dataDir;
  protected int height;
  protected int generation;
  protected int c;

  public PersistentTree(int fanout, int generation, int level, int c, String dataDir) {
    this.fanout = fanout;
    this.generation = generation;
    this.level = level;
    this.c = c;
    this.dataDir = dataDir;
  }

  public abstract void merge(PersistentTree[] trees, int bufferSize);

  public abstract void createTree(List<Entry> entries);

  protected String fileLocation() {
    return String.format("%sL%d/%d/", dataDir, level, generation);
  }

  public synchronized String getLeafFile() {
    return getLevelFileName(getLeafLevel());
  }

  public synchronized String getRootFile() {
    return String.format("%sC%d_root", fileLocation(), c);
  }

  public String getLevelFileName(int level) {
    return String.format("%sC%d_%d", fileLocation(), c, level);
  }

  public int getLeafLevel() {
    return height - 1;
  }

  public static void flushBuffer1(FileOutputStream output, ByteBuffer buffer) throws IOException {
    FileChannel fc = output.getChannel();
    buffer.flip();
    fc.write(buffer);

    buffer.compact();
  }

  protected static int parentNodesNeeded(long children, int fanout) {
    return (int) (children + fanout - 1) / fanout;
  }

  protected static int calcHeight(long n, long b) {
    return (int) Math.ceil(Math.log(n) / Math.log(b));
  }

  protected static int nodesAtGivenLevel(long numberOfLeaves, int fanout, int level) {
    return (int) Math.ceil(numberOfLeaves / Math.pow(fanout, calcHeight(numberOfLeaves, fanout) - (level + 1)));
  }

  protected static boolean isAllEmpty(LinkedList<Entry>[] entryBuffers, int n) {
    for (int i = 0; i < n; i++) {
      if (!entryBuffers[i].isEmpty()) {
        return false;
      }
    }
    return true;
  }

  protected static Entry getNext(LinkedList<Entry>[] entryBuffers, int n) {
    int lowestPos;
    for (lowestPos = 0; lowestPos < n; lowestPos++) {
      if (!entryBuffers[lowestPos].isEmpty()) {
        break;
      }
    }
    //Entry next;
    for (int i = lowestPos; i < n; i++) {
      if (!entryBuffers[i].isEmpty()) {
        if (Long.compareUnsigned(entryBuffers[i].getFirst().lhv, entryBuffers[lowestPos].getFirst().lhv) < 0) {
          lowestPos = i;
        }
      }

    }
    return entryBuffers[lowestPos].pollFirst();
  }

  protected static boolean fillEntryBuffers(int n, LinkedList<Entry>[] entryBuffers, FileInputStream[] streams, ByteBuffer[] buffers, int[] inputStatus) throws IOException {

    for (int i = 0; i < n; i++) {
      if (entryBuffers[i].isEmpty() && inputStatus[i] > 0) {
        inputStatus[i] = IOLSM.fillBuffer1(entryBuffers[i], streams[i], buffers[i]);
      }
    }

    boolean status = true;
    for (int i = 0; i < n; i++) {
      if (inputStatus[i] > 0) {
        status = false;
      }
    }
    return status;
  }

}
