package no.vestein.rtree.lsm;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.LinkedList;
import java.util.List;

public class StreamPersistentTree extends PersistentTree {

  public StreamPersistentTree(int fanout, int generation, int level, int c, String dataDir) {
    super(fanout, generation, level, c, dataDir);
  }

  @Override
  public synchronized void merge(PersistentTree[] trees, int bufferSize) {
    int n = trees.length;

    IOLSM.NodeFileHeader entryHeader;

    FileInputStream[] streams = new FileInputStream[n];
    int[] inputStatus = new int[n];
    IOLSM.NodeFileHeader[] headers = new IOLSM.NodeFileHeader[n];
    ByteBuffer[] readBuffers = new ByteBuffer[n];
    LinkedList<Entry>[] entryBuffer = new LinkedList[n];

    // Output for leaves
//    ByteBuffer writeBuffer = ByteBuffer.allocate(bufferSize * IOLSM.ENTRY_SIZE);
    ByteBuffer writeBuffer = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
    FileOutputStream output;

    // Output buffers for node levels
    ByteBuffer[] nodeWriteBuffers;
    FileOutputStream[] nodeOutputs;
    Node[] tempNodes;
    int[][] parentCounters;
    IOLSM.NodeFileHeader[] nodeOutputHeaders;

    try {
      long totalSize = 0;
      for (int i = 0; i < n; i++) {
        streams[i] = new FileInputStream(trees[i].getLeafFile());
        headers[i] = IOLSM.bytesToNodeFileHeader(streams[i].readNBytes(IOLSM.HEADER_SIZE));

//        readBuffers[i] = ByteBuffer.allocate(bufferSize * IOLSM.ENTRY_SIZE);
        readBuffers[i] = ByteBuffer.allocate(IOLSM.READ_BUFFER_SIZE);
        entryBuffer[i] = new LinkedList<>();
        inputStatus[i] = 1;

        totalSize += headers[i].size;
      }

      // Expected tree height
      height = calcHeight(totalSize, fanout);

      // Output initialization
      nodeWriteBuffers = new ByteBuffer[height];
      nodeOutputs = new FileOutputStream[height];
      tempNodes = new Node[height];
      parentCounters = new int[height][2];
      nodeOutputHeaders = new IOLSM.NodeFileHeader[height];


      for (int i = height - 1; i > 0; i--) {
        nodeWriteBuffers[i] = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
        nodeOutputs[i] = new FileOutputStream(getLevelFileName(i - 1));
        parentCounters[i][0] = fanout;
        parentCounters[i][1] = IOLSM.HEADER_SIZE;

//        tempNodes[i] = new Node(Integer.MAX_VALUE, Integer.MIN_VALUE, Integer.MAX_VALUE, Integer.MIN_VALUE);
        tempNodes[i] = new Node(Float.MAX_VALUE, -Float.MAX_VALUE, Float.MAX_VALUE, -Float.MAX_VALUE);
        nodeOutputHeaders[i] = new IOLSM.NodeFileHeader(nodesAtGivenLevel(totalSize, fanout, i - 1), IOLSM.NODE_SIZE, i - 1, IOLSM.NodeFileHeader.NODE);
        nodeOutputs[i].write(IOLSM.nodeFileHeaderToBytes(nodeOutputHeaders[i]));
      }

      // Prepare root file
      nodeWriteBuffers[0] = ByteBuffer.allocate(IOLSM.NODE_SIZE);
      nodeOutputs[0] = new FileOutputStream(getRootFile());
      parentCounters[0][0] = fanout;
//      tempNodes[0] = new Node(Integer.MAX_VALUE, Integer.MIN_VALUE, Integer.MAX_VALUE, Integer.MIN_VALUE);
      tempNodes[0] = new Node(Float.MAX_VALUE, -Float.MAX_VALUE, Float.MAX_VALUE, -Float.MAX_VALUE);
      nodeOutputHeaders[0] = new IOLSM.NodeFileHeader(1, IOLSM.NODE_SIZE, -1, IOLSM.NodeFileHeader.ROOT);
      nodeOutputs[0].write(IOLSM.nodeFileHeaderToBytes(nodeOutputHeaders[0]));

      // Prepare leaf file
      entryHeader = new IOLSM.NodeFileHeader(totalSize, IOLSM.ENTRY_SIZE, height - 1, IOLSM.NodeFileHeader.ENTRY);
      output = new FileOutputStream(getLevelFileName(height - 1));
      output.write(IOLSM.nodeFileHeaderToBytes(entryHeader));

      // Merge input files
      fillEntryBuffers(n, entryBuffer, streams, readBuffers, inputStatus);

      while (!isAllEmpty(entryBuffer, n)) {
        Entry next = getNext(entryBuffer, n);
        updateParents(next, height - 1, nodeWriteBuffers, nodeOutputs, tempNodes, parentCounters);

        writeBuffer.put(IOLSM.entryToBytes(next));
        if (writeBuffer.remaining() < IOLSM.ENTRY_SIZE) {
          flushBuffer1(output, writeBuffer);
        }

        boolean status = fillEntryBuffers(n, entryBuffer, streams, readBuffers, inputStatus);
      }

      // Flush entry level
      flushBuffer1(output, writeBuffer);
      output.close();

      // Close input streams
      for (int i = 0; i < n; i++) {
        streams[i].close();
      }

      // Force update upwards
      for (int i = height - 2; i > 0; i--) {
        tempNodes[i].updateDimension(tempNodes[i + 1]);
        tempNodes[i].updateLhv(tempNodes[i + 1].lhv);
        tempNodes[i].length += IOLSM.NODE_SIZE;
      }

      // Flush node levels
      for (int i = 0; i < height; i++) {
        nodeWriteBuffers[i].put(IOLSM.nodeToBytes(tempNodes[i]));
        flushBuffer1(nodeOutputs[i], nodeWriteBuffers[i]);
        nodeOutputs[i].close();
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  @Override
  public synchronized void createTree(List<Entry> entries) {
    IOLSM.NodeFileHeader entryHeader;

    ByteBuffer writeBuffer = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
    FileOutputStream output;

    ByteBuffer[] nodeWriteBuffers;
    FileOutputStream[] nodeOutputs;
    Node[] tempNodes;
    int[][] parentCounters;
    IOLSM.NodeFileHeader[] nodeOutputHeaders;

    try {
      long totalSize = entries.size();

      height = calcHeight(totalSize, fanout);

      // Output initialization
      nodeWriteBuffers = new ByteBuffer[height];
      nodeOutputs = new FileOutputStream[height];
      tempNodes = new Node[height];
      parentCounters = new int[height][2];
      nodeOutputHeaders = new IOLSM.NodeFileHeader[height];

      for (int i = height - 1; i > 0; i--) {
        nodeWriteBuffers[i] = ByteBuffer.allocate(IOLSM.WRITE_BUFFER_SIZE);
        nodeOutputs[i] = new FileOutputStream(getLevelFileName(i - 1));
        parentCounters[i][0] = fanout;
        parentCounters[i][1] = IOLSM.HEADER_SIZE;
        tempNodes[i] = new Node(Float.MAX_VALUE, -Float.MAX_VALUE, Float.MAX_VALUE, -Float.MAX_VALUE);
        nodeOutputHeaders[i] = new IOLSM.NodeFileHeader(nodesAtGivenLevel(totalSize, fanout, i - 1), IOLSM.NODE_SIZE, i - 1, IOLSM.NodeFileHeader.NODE);
        nodeOutputs[i].write(IOLSM.nodeFileHeaderToBytes(nodeOutputHeaders[i]));
      }

      // Prepare root file
      nodeWriteBuffers[0] = ByteBuffer.allocate(IOLSM.NODE_SIZE);
      nodeOutputs[0] = new FileOutputStream(getRootFile());
      parentCounters[0][0] = fanout;
//      tempNodes[0] = new Node(Integer.MAX_VALUE, Integer.MIN_VALUE, Integer.MAX_VALUE, Integer.MIN_VALUE);
      tempNodes[0] = new Node(Float.MAX_VALUE, -Float.MAX_VALUE, Float.MAX_VALUE, -Float.MAX_VALUE);
      nodeOutputHeaders[0] = new IOLSM.NodeFileHeader(1, IOLSM.NODE_SIZE, -1, IOLSM.NodeFileHeader.ROOT);
      nodeOutputs[0].write(IOLSM.nodeFileHeaderToBytes(nodeOutputHeaders[0]));

      // Prepare leaf file
      entryHeader = new IOLSM.NodeFileHeader(totalSize, IOLSM.ENTRY_SIZE, height - 1, IOLSM.NodeFileHeader.ENTRY);
      output = new FileOutputStream(getLevelFileName(height - 1));
      output.write(IOLSM.nodeFileHeaderToBytes(entryHeader));

      Entry next;
      for (int i = 0; i < totalSize; i++) {
        next = entries.get(i);
        updateParents(next, height - 1, nodeWriteBuffers, nodeOutputs, tempNodes, parentCounters);

        writeBuffer.put(IOLSM.entryToBytes(next));
        if (writeBuffer.remaining() < IOLSM.ENTRY_SIZE) {
          flushBuffer1(output, writeBuffer);
        }
      }

      // Flush leaf level
      flushBuffer1(output, writeBuffer);
      output.close();

      // Force update upwards
      for (int i = height - 2; i > 0; i--) {
        tempNodes[i].updateDimension(tempNodes[i + 1]);
        tempNodes[i].updateLhv(tempNodes[i + 1].lhv);
        tempNodes[i].length += IOLSM.NODE_SIZE;
      }

      // Flush node levels
      for (int i = 0; i < height; i++) {
        nodeWriteBuffers[i].put(IOLSM.nodeToBytes(tempNodes[i]));
        flushBuffer1(nodeOutputs[i], nodeWriteBuffers[i]);
        nodeOutputs[i].close();
      }

    } catch (Exception e) {
      e.printStackTrace();
    }

  }

  // fix this
  private void updateParents(Entry entry, int level, ByteBuffer[] nodeWriteBuffers, FileOutputStream[] nodeOutputs, Node[] tempNodes, int[][] parentCounters) throws IOException {

    if (parentCounters[level][0] > 0) {
      tempNodes[level].updateDimension(entry);
      tempNodes[level].updateLhv(entry.lhv);
      tempNodes[level].length += IOLSM.ENTRY_SIZE;
      parentCounters[level][0]--;
      return;
    }

    nodeWriteBuffers[level].put(IOLSM.nodeToBytes(tempNodes[level]));
    if (nodeWriteBuffers[level].remaining() < IOLSM.NODE_SIZE) {
      flushBuffer1(nodeOutputs[level], nodeWriteBuffers[level]);
    }

    int ups = level;
    while (--ups >= 0) {
      if (parentCounters[ups][0] > 0) {
        tempNodes[ups].updateDimension(tempNodes[ups + 1]);
        tempNodes[ups].updateLhv(tempNodes[ups + 1].lhv);
        tempNodes[ups].length += IOLSM.NODE_SIZE;
        parentCounters[ups][0]--;
        break;
      }

      nodeWriteBuffers[ups].put(IOLSM.nodeToBytes(tempNodes[ups]));
      if (nodeWriteBuffers[ups].remaining() < IOLSM.NODE_SIZE) {
        flushBuffer1(nodeOutputs[ups], nodeWriteBuffers[ups]);
      }
    }

    parentCounters[level][0] = fanout - 1;
    tempNodes[level].setData(entry);
    tempNodes[level].start += tempNodes[level].length;
    tempNodes[level].length = IOLSM.ENTRY_SIZE;

    while (++ups < level) {
      parentCounters[ups][0] = fanout - 1;
      tempNodes[ups].setData(tempNodes[ups + 1]);
      tempNodes[ups].start += tempNodes[ups].length;
      tempNodes[ups].length = IOLSM.NODE_SIZE;
    }

  }



}
